# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os
from os.path import basename, dirname
from functools import wraps

from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory
from mvc.support.gui_loop import stop_event_loop, add_idle_call

from pyxrd.data import settings

from pyxrd.generic.controllers import BaseController
from pyxrd.generic.plot.controllers import MainPlotController
from pyxrd.generic.plot.eye_dropper import EyeDropper

from pyxrd.file_parsers.project_parsers import JSONProjectParser
from pyxrd.file_parsers.project_parsers import project_parsers

from pyxrd.project.controllers import ProjectController
from pyxrd.project.models import Project
from pyxrd.specimen.controllers import SpecimenController, MarkersController

from pyxrd.mixture.controllers import MixturesController #, InSituBehavioursController
from pyxrd.phases.controllers import PhasesController
from pyxrd.atoms.controllers import AtomTypesController

class AppController (BaseController):
    """
        Controller handling the main application interface.
        In essence this delegates actions to its child controllers for Project,
        Mixture, Specimen, Phase, Marker and Atoms actions. 
    """

    # ------------------------------------------------------------
    #      Dialog properties
    # ------------------------------------------------------------
    _save_project_dialog = None
    @property
    def save_project_dialog(self):
        """ Creates & returns the 'save project' dialog """
        if self._save_project_dialog is None:
            # Check to see if we have a project loaded, if so,
            # set the paths to match
            current_name, current_folder = None, None
            if self.model.current_filename is not None:
                current_name = basename(self.model.current_filename)
                current_folder = dirname(self.model.current_filename)
            # Create the dialog once, and re-use its context
            self._save_project_dialog = DialogFactory.get_save_dialog(
                title="Save project",
                current_name=current_name,
                current_folder=current_folder,
                filters=project_parsers.get_export_file_filters(),
                persist=True,
                parent=self.view.get_top_widget()
            )
        return self._save_project_dialog

    _load_project_dialog = None
    @property
    def load_project_dialog(self):
        """ Creates & returns the 'load project' dialog """
        if self._load_project_dialog is None:
            # Check to see if we have a project loaded, if so,
            # set the paths to match
            current_folder = None
            if self.model.current_filename is not None:
                current_folder = dirname(self.model.current_filename)
            # Create the dialog once, and re-use
            self._load_project_dialog = DialogFactory.get_load_dialog(
                title="Load project",
                current_folder=current_folder,
                filters=project_parsers.get_import_file_filters(),
                persist=True, multiple=False,
                parent=self.view.get_top_widget()
            )
        return self._load_project_dialog

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, model, view, gtk_exception_hook=None, spurious=False, auto_adapt=False, parent=None):
        """ Initializes an AppController with the given arguments. """
        super(AppController, self).__init__(model=model, view=view, spurious=spurious, auto_adapt=auto_adapt, parent=parent)

        self.gtk_exception_hook = gtk_exception_hook
        self.gtk_exception_hook.parent_view = view.get_toplevel()

        # Plot controller:
        self.plot_controller = MainPlotController(
            self.update_plot_status,
            self.show_marker
        )

        # Child controllers:
        self.project = None
        self.specimen = None
        self.markers = None
        self.phases = None
        self.atom_types = None
        self.mixtures = None

        self.idle_redraw_plot()

        if self.model.project_loaded:
            self.reset_project_controller()

        self.push_status_msg("Done.")

    def register_view(self, view):
        """ Registers the view with this controller """
        view.setup_plot(self.plot_controller)
        if self.model.project_loaded:
            self.update_sensitivities()
            view.set_layout_mode(self.model.current_project.layout_mode)
        else:
            view.set_layout_mode(settings.DEFAULT_LAYOUT)

    def set_model(self, model):
        """ Sets the model in this controller """
        super(self, AppController).set_model(model)
        self.reset_project_controller()

    def reset_project_controller(self):
        """ Recreates all child controllers """
        self.view.reset_all_views()
        self.project = ProjectController(model=self.model.current_project, view=self.view.project, parent=self)
        self.phases = PhasesController(model=self.model.current_project, view=self.view.phases, parent=self)
        #self.behaviours = InSituBehavioursController(model=self.model.current_project, view=self.view.behaviours, parent=self)
        self.atom_types = AtomTypesController(model=self.model.current_project, view=self.view.atom_types, parent=self)
        self.mixtures = MixturesController(model=self.model.current_project, view=self.view.mixtures, parent=self)
        self.reset_specimen_controller()
        self.view.update_project_sensitivities(self.model.project_loaded)
        self.set_layout_mode(self.model.current_project.layout_mode)
        self.update_title()

    def reset_specimen_controller(self):
        """ Recreates only the specimen controllers """
        if self.model.specimen_selected:
            specimen_view = self.view.reset_child_view("specimen")
            self.specimen = SpecimenController(model=self.model.current_specimen, view=specimen_view, parent=self)
            self.show_markers_for(self.model.current_specimen)
        else:
            self.specimen = None
            self.markers = None
        self.view.update_specimen_sensitivities(
            self.model.single_specimen_selected,
            self.model.multiple_specimens_selected
        )
        self.idle_redraw_plot()

    def show_markers_for(self, specimen):
        markers_view = self.view.reset_child_view("markers")
        self.markers = MarkersController(model=specimen, view=markers_view, parent=self)
        
    def show_marker(self, marker):
        markers_view = self.view.reset_child_view("markers")
        self.markers = MarkersController(model=marker.specimen, view=markers_view, parent=self)
        self.view.markers.present()
        add_idle_call(self.markers.select_object, marker)
             

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @BaseController.observe("needs_plot_update", signal=True)
    def notif_needs_plot_update(self, model, prop_name, info):
        """ 
            This handles needs_plot_update signals emitted by the Application
            model, in effect it is either a forwarded 'data_changed' or 
            'visuals_changed' signal coming from the 
            :class:`pyxrd.project.models.Project` model.
        """
        self.idle_redraw_plot()

    @BaseController.observe("current_project", assign=True, after=True)
    def notif_project_update(self, model, prop_name, info):
        self.reset_project_controller()

    @BaseController.observe("current_specimen", assign=True, after=True)
    @BaseController.observe("current_specimens", assign=True, after=True)
    def notif_specimen_changed(self, model, prop_name, info):
        self.reset_specimen_controller()

    # ------------------------------------------------------------
    #      View updating
    # ------------------------------------------------------------

    _idle_redraw_id = None
    _needs_redraw = False
    def idle_redraw_plot(self):
        """Adds a redraw plot function as 'idle' action to the main GTK loop."""
        if self._idle_redraw_id is None:
            self._idle_redraw_id = add_idle_call(self.redraw_plot)
        self._needs_redraw = True

    @BaseController.status_message("Updating display...")
    def redraw_plot(self):
        """Updates the plot"""
        if self._needs_redraw == True:
            self._needs_redraw = False
            self.plot_controller.update(
                clear=True,
                project=self.model.current_project,
                specimens=self.model.current_specimens[::-1]
            )
        if self._needs_redraw:
            return True
        else:
            self._idle_redraw_id = None
            return False


    def update_title(self):
        """Convenience method for setting the application view's title"""
        self.view.set_title(self.model.current_project.name)

    def update_sensitivities(self):
        """Convenience method for updating the application view's sensitivities"""
        self.view.update_project_sensitivities(self.model.project_loaded)
        self.view.update_specimen_sensitivities(
            self.model.single_specimen_selected,
            self.model.multiple_specimens_selected
        )

    def update_plot_status(self, x_pos, event):
        if x_pos > 0 and self.model.current_specimen is not None:
            # Get experimental data at the sampled point
            exp_y = self.model.current_specimen.experimental_pattern.get_y_at_x(x_pos)
            dspacing = self.model.current_specimen.goniometer.get_nm_from_2t(x_pos)
            # Get calculated data if applicable
            if self.model.current_project.layout_mode == "FULL":
                calc_y = self.model.current_specimen.calculated_pattern.get_y_at_x(x_pos)
                self.view.update_plot_status(x_pos, dspacing, exp_y, calc_y)
            else:
                self.view.update_plot_status(x_pos, dspacing, exp_y, None)
        else:
            self.view.update_plot_status(None, None, None, None)

    def set_layout_mode(self, mode):
        """Convenience method for updating the application view's layout mode"""
        self.view.set_layout_mode(mode)

    # ------------------------------------------------------------
    #      Loading and saving of projects
    # ------------------------------------------------------------
    def _save_project(self, filename=None):
        # Set the filename to the current location if None or "" was given:
        filename = filename or self.model.current_filename

        # Try to save the project:
        with DialogFactory.error_dialog_handler(
                "An error has occurred while saving!\n<i>{0}</i>",
                parent=self.view.get_toplevel(), reraise=False):
            JSONProjectParser.write(self.model.current_project, filename, zipped=True)
            self.model.current_project.filename = filename
            self.model.update_project_last_save_hash()
            
        # Update the title
        self.update_title()

    def confirm_discard_unsaved_changes(self,
            confirm_msg="The current project has unsaved changes,\n"
                        "are you sure you want to continue?",
            on_reject=None):
        """
            Function decorator which will check if a project is opened with
            unsaved changes and ask the user to confirm the action without first
            saving the changes.
        """
        def accept_decorator(on_accept):
            @wraps(on_accept)
            def accept_wrapper(self, *args, **kwargs):
                if self.model.check_for_changes():
                    return DialogFactory.get_confirmation_dialog(
                        confirm_msg, parent=self.view.get_top_widget()
                    ).run(lambda d: on_accept(self, *args, **kwargs), on_reject)
                else:
                    return on_accept(self, *args, **kwargs)
            return accept_wrapper
        return accept_decorator

    @confirm_discard_unsaved_changes(
        "The current project has unsaved changes,\n"
        "are you sure you want to quit?")
    def quit(self, *args, **kwargs):
        stop_event_loop()
        return False

    @confirm_discard_unsaved_changes(
        "The current project has unsaved changes,\n"
        "are you sure you want to load another project?")
    def load_project(self):
        """Convenience function for loading projects from different sources
        following similar user interaction paths"""
        def on_accept(dialog):
            # Try to load the project:
            with DialogFactory.error_dialog_handler(
                    "An error has occurred:\n<i>{0}</i>\n Your project was not loaded!",
                    parent=self.view.get_toplevel(), title="Parsing error", reraise=False):
                self.model.current_project = dialog.parser.parse(dialog.filename)
                self.model.current_project.parent = self.model
                self.model.update_project_last_save_hash()
                # Update the title
                self.update_title()

        # Run the open/import project dialog:
        self.load_project_dialog.run(on_accept)

    @confirm_discard_unsaved_changes(
        "The current project has unsaved changes,\n"
        "are you sure you want to create a new project?")
    def new_project(self, *args, **kwargs):
        # Create a new project
        self.model.current_project = Project(parent=self.model)

        # Set the current filename property and update the title
        self.update_title()

        # Show the edit project dialog
        self.view.project.present()

    # ------------------------------------------------------------
    #      GTK Signal handlers - general
    # ------------------------------------------------------------
    def on_manual_activate(self, widget, data=None):
        try:
            import webbrowser
            webbrowser.open(settings.MANUAL_URL)
        except:
            pass # ignore errors
        return True

    def on_about_activate(self, widget, data=None):
        self.view["about_window"].show()
        return True

    def on_main_window_destroy_event(self, widget, event):
        self.quit()
        return True

    def on_main_window_delete_event(self, widget, event):
        self.quit()
        return True

    def on_menu_item_quit_activate (self, widget, data=None):
        self.quit()
        return True

    def on_refresh_graph(self, event):
        if self.model.current_project:
            with self.model.current_project.data_changed.hold():
                self.model.current_project.update_all_mixtures()
                self.redraw_plot()

    def on_save_graph(self, event):
        filename = None
        if self.model.single_specimen_selected:
            filename = os.path.splitext(self.model.current_specimen.name)[0]
        else:
            filename = self.model.current_project.name
        self.plot_controller.save(
            parent=self.view.get_toplevel(),
            current_name=filename,
            num_specimens=len(self.model.current_specimens),
            offset=self.model.current_project.display_plot_offset)

    def on_toggled_plot_toolbar(self, action):
        if action.get_active():
            self.view.show_plot_toolbar()
        else:
            self.view.hide_plot_toolbar()

    @BaseController.status_message("Sampling...", "sampling")
    def on_sample_point(self, event):
        """
            Sample a point on the plot and display the (calculated and)
            experimental data values in an information dialog.
        """

        self.edc = None

        def parse_x_pos(x_pos, event):
            # Clear the eye dropper controller
            self.edc.enabled = False
            self.edc.disconnect()
            del self.edc
            # Get experimental data at the sampled point
            exp_y = self.model.current_specimen.experimental_pattern.get_y_at_x(x_pos)
            message = "Sampled point:\n"
            message += "\tExperimental data:\t( %.4f , %.4f )\n" % (x_pos, exp_y)
            # Get calculated data if applicable
            if self.model.current_project.layout_mode == "FULL":
                calc_y = self.model.current_specimen.calculated_pattern.get_y_at_x(x_pos)
                message += "\tCalculated data:\t\t( %.4f , %.4f )" % (x_pos, calc_y)
            # Display message dialog
            DialogFactory.get_information_dialog(
                message, parent=self.view.get_toplevel()
            ).run()


        self.edc = EyeDropper(self.plot_controller, parse_x_pos)

    # ------------------------------------------------------------
    #      GTK Signal handlers - Project related
    # ------------------------------------------------------------
    @BaseController.status_message("Creating new project...", "new_project")
    def on_new_project_activate(self, widget, data=None):
        self.new_project()

    @BaseController.status_message("Displaying project data...", "edit_project")
    def on_edit_project_activate(self, widget, data=None):
        self.view.project.present()

    @BaseController.status_message("Open project...", "open_project")
    def on_open_project_activate(self, widget, data=None):
        """Open an existing project. Asks the user if (s)he's sure when an 
        unsaved project is loaded."""
        self.load_project()

    @BaseController.status_message("Save project...", "save_project")
    def on_save_project_activate(self, widget, *args):
        # No filename yet: show a dialog
        if not self.model.current_filename:
            self.save_project_dialog.update(title="Save project").run(
                lambda dialog: self._save_project(filename=dialog.filename)
            )
        else: # we already have a filename, overwrite:
            self._save_project()

    @BaseController.status_message("Save project as...", "save_project_as")
    def on_save_project_as_activate(self, widget, *args):
        self.save_project_dialog.update(title="Save project as").run(
            lambda dialog: self._save_project(filename=dialog.filename)
        )

    # ------------------------------------------------------------
    #      GTK Signal handlers - Mixtures related
    # -----------------------------------------------------------
    def on_edit_mixtures(self, widget, data=None):
        if self.model.project_loaded:
            self.view.mixtures.present()
        pass

    # ------------------------------------------------------------
    #      GTK Signal handlers - Behaviour related
    # ------------------------------------------------------------
    def on_edit_behaviours(self, widget, data=None):
        #if self.model.project_loaded:
        #    self.view.behaviours.present()
        return True

    # ------------------------------------------------------------
    #      GTK Signal handlers - Specimen related
    # ------------------------------------------------------------
    def on_edit_specimen_activate(self, event):
        self.project.edit_specimen()
        return True

    def on_add_specimen_activate(self, event):
        self.project.add_specimen()
        return True

    def on_add_multiple_specimens(self, event):
        self.project.import_multiple_specimen()
        return True

    def on_del_specimen_activate(self, event):
        self.project.delete_selected_specimens()
        return True

    def on_replace_specimen_data_activate(self, event):
        if self.model.single_specimen_selected:
            self.specimen.on_replace_experimental_data()
        return True

    def on_export_specimen_data_activate(self, event):
        if self.model.single_specimen_selected:
            self.specimen.on_export_experimental_data()
        return True

    def on_convert_to_fixed_activate(self, event):
        for specimen in self.model.current_specimens:
            specimen.convert_to_fixed()

    def on_convert_to_ads_activate(self, event):
        for specimen in self.model.current_specimens:
            specimen.convert_to_ads()

    def on_remove_background(self, event):
        if self.model.single_specimen_selected:
            self.specimen.remove_background()
        else:
            self.project.remove_backgrounds(self.model.current_specimens)
        return True

    def on_smooth_data(self, event):
        if self.model.single_specimen_selected:
            self.specimen.smooth_data()
        return True

    def on_add_noise(self, event):
        if self.model.single_specimen_selected:
            self.specimen.add_noise()
        return True

    def on_shift_data(self, event):
        if self.model.single_specimen_selected:
            self.specimen.shift_data()
        return True

    def on_strip_peak(self, event):
        if self.model.single_specimen_selected:
            self.specimen.strip_peak()
        return True

    def on_peak_properties(self, event):
        if self.model.single_specimen_selected:
            self.specimen.peak_properties()
        return True

    # ------------------------------------------------------------
    #      GTK Signal handlers - Phases related
    # ------------------------------------------------------------
    def on_edit_phases_activate(self, event):
        if self.model.project_loaded:
            self.view.phases.present()
        return True

    # ------------------------------------------------------------
    #      GTK Signal handlers - Atom Types related
    # ------------------------------------------------------------
    def on_edit_atom_types_activate(self, event):
        if self.model.project_loaded:
            self.view.atom_types.present()
        return True

    # ------------------------------------------------------------
    #      GTK Signal handlers - Markers related
    # ------------------------------------------------------------
    def on_edit_markers_activate(self, event):
        if self.model.current_specimen is not None:
            self.view.markers.present()
        return True

    def edit_marker(self, marker):
        self.show_markers_for(marker.specimen)

    pass # end of class




# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from contextlib import contextmanager

import logging
logger = logging.getLogger(__name__)

from mvc import Model
from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory

from pyxrd.generic.views.treeview_tools import new_text_column
from pyxrd.generic.controllers import ObjectListStoreController
from pyxrd.generic.utils import not_none

from pyxrd.phases.views import (
    EditPhaseView, AddPhaseView, EditRawPatternPhaseView
)

from pyxrd.file_parsers.json_parser import JSONParser
from pyxrd.file_parsers.phase_parsers import phase_parsers

from ..models import Phase, RawPatternPhase

from .edit_phase_controller import EditPhaseController
from .raw_pattern_phase_controller import EditRawPatternPhaseController
from .add_phase_controller import AddPhaseController


class PhasesController(ObjectListStoreController):
    """ 
        Controller for the phases list
    """
    file_filters = Phase.Meta.file_filters
    treemodel_property_name = "phases"
    treemodel_class_type = Phase
    obj_type_map = [
        (Phase, EditPhaseView, EditPhaseController),
        (RawPatternPhase, EditRawPatternPhaseView, EditRawPatternPhaseController),
    ]
    multi_selection = True
    columns = [
        ("Phase name", "c_name"),
        (" ", "c_display_color"),
        ("R", "c_R"),
        ("#", "c_G"),
    ]
    delete_msg = "Deleting a phase is irreversible!\nAre You sure you want to continue?"
    title = "Edit Phases"

    def get_phases_tree_model(self, *args):
        return self.treemodel

    def load_phases(self, filename, parser=JSONParser):
        index = self.get_selected_index()
        if index is not None: index += 1
        self.model.load_phases(filename, parser=parser, insert_index=index)

    def setup_treeview_col_c_display_color(self, treeview, name, col_descr, col_index, tv_col_nr):
        def set_background(column, cell_renderer, tree_model, iter, col_index):
            try:
                color = tree_model.get_value(iter, col_index)
            except TypeError:
                pass  # invalid iter
            else:
                cell_renderer.set_property('background', color)
                cell_renderer.set_property('text', "")

        treeview.append_column(new_text_column(
            name,
            data_func=(set_background, (col_index,)),
            text_col=col_index,
            resizable=False,
            expand=False))

        return True

    def create_new_object_proxy(self):
        def on_accept(phase_type, G, R):
            index = int(not_none(self.get_selected_index(), -1)) + 1
            if phase_type == "empty":
                self.add_object(Phase(G=int(G), R=int(R)))
            elif phase_type == "raw":
                self.add_object(RawPatternPhase())
            else:
                filename = phase_type
                if filename != None:
                    self.model.load_phases(filename, parser=JSONParser, insert_index=index)

        # TODO re-use this and reset the COMBO etc.
        self.add_model = Model()
        self.add_view = AddPhaseView(parent=self.view)
        self.add_ctrl = AddPhaseController(
            model=self.add_model, view=self.add_view, parent=self.parent,
            callback=on_accept
        )

        self.add_view.present()
        return None

    @contextmanager
    def _multi_operation_context(self):
        with self.model.hold_mixtures_data_changed():
            with self.model.hold_mixtures_needs_update():
                with self.model.hold_phases_data_changed():
                    yield

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_save_object_clicked(self, event):
        def on_accept(dialog):
            logger.info("Exporting phases...")
            Phase.save_phases(self.get_selected_objects(), filename=dialog.filename)
        DialogFactory.get_save_dialog(
            "Export phase", parent=self.view.get_top_widget(),
            filters=phase_parsers.get_export_file_filters()
        ).run(on_accept)
        return True


    def on_load_object_clicked(self, event):
        def on_accept(dialog):
            logger.info("Importing phases...")
            self.load_phases(dialog.filename, parser=dialog.parser)
        DialogFactory.get_load_dialog(
            "Import phase", parent=self.view.get_top_widget(),
            filters=phase_parsers.get_import_file_filters()
        ).run(on_accept)
        return True



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

from mvc import Controller
from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory

from pyxrd.generic.controllers import BaseController
from pyxrd.generic.controllers.objectliststore_controllers import wrap_list_property_to_treemodel

from pyxrd.refinement.views.refinement_view import RefinementView
from pyxrd.refinement.controllers.refinement_controller import RefinementController

class EditMixtureController(BaseController):

    auto_adapt_excluded = [
        "refine_method_index",
        "refinables",
        "make_psp_plots"
    ]

    ref_view = None

    @property
    def specimens_treemodel(self):
        if self.model.project is not None:
            return wrap_list_property_to_treemodel(self.model.project, type(self.model.project).specimens)
        else:
            return None

    @property
    def phases_treemodel(self):
        if self.model.project is not None:
            return wrap_list_property_to_treemodel(self.model.project, type(self.model.project).phases)
        else:
            return None

    def register_adapters(self):
        self.create_ui()

    def create_ui(self):
        """
            Creates a complete new UI for the Mixture model
        """
        self.view.reset_view()
        for index in range(len(self.model.phases)):
            self._add_phase_view(index)
        for index in range(len(self.model.specimens)):
            self._add_specimen_view(index)

    def _add_phase_view(self, phase_slot):
        """
            Adds a new view for the given phase slot.
        """
        def on_label_changed(editable):
            self.model.phases[phase_slot] = editable.get_text()

        def on_check_toggled(check):
            self.model.fractions_mask[phase_slot] = 1 if check.get_active() else 0

        def on_fraction_changed(editable):
            try: self.model.fractions[phase_slot] = float(editable.get_text())
            except ValueError: return # ignore ValueErrors

        def on_phase_delete(widget):
            self.model.del_phase_slot(phase_slot)
            widget.disconnect(getattr(widget, "deleventid"))

        self.view.add_phase_slot(self.phases_treemodel,
            on_phase_delete, on_label_changed, on_check_toggled, on_fraction_changed,
            self.on_combo_changed, label=self.model.phases[phase_slot],
            fraction=self.model.fractions[phase_slot], phases=self.model.phase_matrix)

    def _add_specimen_view(self, specimen_slot):
        """
            Adds a new view for the given specimen slot
        """
        def on_scale_changed(editable):
            try: self.model.scales[specimen_slot] = float(editable.get_text())
            except ValueError: return # ignore ValueErrors

        def on_bgs_changed(editable):
            try: self.model.bgshifts[specimen_slot] = float(editable.get_text())
            except ValueError: return # ignore ValueErrors

        def on_specimen_changed(combobox):
            itr = combobox.get_active_iter()
            specimen = self.specimens_treemodel.get_user_data(itr) if itr is not None else None
            self.model.set_specimen(specimen_slot, specimen)

        def on_specimen_delete(widget):
            self.model.del_specimen_slot(specimen_slot)
            widget.disconnect(getattr(widget, "deleventid"))

        self.view.add_specimen_slot(self.phases_treemodel,
            self.specimens_treemodel, on_specimen_delete, on_scale_changed,
            on_bgs_changed, on_specimen_changed, self.on_combo_changed,
            scale=self.model.scales[specimen_slot], bgs=self.model.bgshifts[specimen_slot],
            specimen=self.model.specimens[specimen_slot], phases=self.model.phase_matrix)

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @Controller.observe("data_changed", signal=True)
    def notif_has_changed(self, model, prop_name, info):
        self.view.update_all(self.model.fractions, self.model.scales, self.model.bgshifts)

    @Controller.observe("needs_reset", signal=True)
    def notif_needs_reset(self, model, prop_name, info):
        self.create_ui()

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_combo_changed(self, combobox, row, col):
        itr = combobox.get_active_iter()
        phase = self.phases_treemodel.get_user_data(itr) if itr is not None else None
        self.model.set_phase(row, col, phase)

    def on_add_phase(self, widget, *args):
        with self.model.data_changed.hold():
            index = self.model.add_phase_slot("New Phase", 1.0)
            self._add_phase_view(index)

    def on_add_specimen(self, widget, *args):
        with self.model.data_changed.hold():
            index = self.model.add_specimen_slot(None, 1.0, 0.0)
            self._add_specimen_view(index)

    def on_add_both(self, widget, *args):
        with self.model.data_changed.hold():
            self.on_add_specimen(widget, *args)
            self.on_add_phase(widget, *args)

    def on_optimize_clicked(self, widget, *args):
        self.model.optimize()

    def on_refine_clicked(self, widget, *args):
        self.model.refinement.update_refinement_treestore()
        if self.ref_view is not None:
            self.ref_view.hide()
            self.ref_ctrl.cleanup()
        self.view.parent.hide()
        self.ref_view = RefinementView(parent=self.parent.view)
        self.ref_ctrl = RefinementController(model=self.model.refinement, view=self.ref_view, parent=self)
        self.ref_view.present()

    def on_composition_clicked(self, widget, *args):
        comp = "The composition of the specimens in this mixture:\n\n\n"
        comp += "<span font-family=\"monospace\">"
        # get the composition matrix (first columns contains strings with elements, others are specimen compositions)
        import re
        for row in self.model.get_composition_matrix():
            comp += "%s %s\n" % (re.sub(r'(\d+)', r'<sub>\1</sub>', row[0]), " ".join(row[1:]))
        comp += "</span>"
        DialogFactory.get_information_dialog(
            comp, parent=self.view.get_toplevel()
        ).run()

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

import numpy as np

from pyxrd.calculations.data_objects import AtomData, AtomTypeData
from pyxrd.calculations.atoms import get_atomic_scattering_factor, get_structure_factor

__all__ = [
    'TestAtomCalcs',
]

class TestAtomCalcs(unittest.TestCase):

    def setUp(self):
        self.atom_type_data = AtomTypeData( # this is the data for a H atom
            par_a = np.asanyarray([0.413048,0.294953,0.187491,0.080701,0.023736]),
            par_b = np.asanyarray([15.569946,32.398468,5.711404,61.889874,1.334118]),
            par_c = 0.000049,
            debye = 0
        )
        self.atom_data = AtomData(
            atom_type = self.atom_type_data,
            pn = 1,
            default_z = 0,
            z = 0
        )       

    def tearDown(self):
        del self.atom_data
        del self.atom_type_data

    def test_not_none(self):
        self.assertIsNotNone(self.atom_type_data)
        self.assertIsNotNone(self.atom_data)

    def test_scattering_factor(self):
        result = get_atomic_scattering_factor(
            np.asanyarray([2.2551711385, 2.478038901, 2.7001518288, 2.9214422642, 3.1418428, 3.3612863, 3.5797059197, 3.7970351263]),
            self.atom_type_data
        )
        self.assertIsNotNone(result)

    def test_structure_factor(self):
        result = get_structure_factor(
            np.asanyarray([2.2551711385, 2.478038901, 2.7001518288, 2.9214422642, 3.1418428, 3.3612863, 3.5797059197, 3.7970351263]),
            self.atom_data
        )
        self.assertIsNotNone(result)

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from mvc import Observer
from pyxrd.generic.models import ChildModel

__all__ = [
    'TestChildModel',
]

class TestChildModel(unittest.TestCase):

    def setUp(self):
        self.childmodel = ChildModel()

        class ChildObserver(Observer):

            added_recieved = False
            @Observer.observe("added", signal=True)
            def notify_added(self, model, prop_name, info):
                self.added_recieved = True

            removed_recieved = False
            @Observer.observe("removed", signal=True)
            def notify_removed(self, model, prop_name, info):
                self.removed_recieved = True

        self.observer = ChildObserver(model=self.childmodel)

    def tearDown(self):
        self.observer.relieve_model(self.childmodel)
        del self.observer
        del self.childmodel

    def test_not_none(self):
        self.assertIsNotNone(self.childmodel)

    def test_signals(self):
        # Reset flags:
        self.observer.removed_recieved = False
        self.observer.added_recieved = False

        # No parent set, only the added signal is fired,
        self.childmodel.parent = object()

        self.assertFalse(self.observer.removed_recieved)
        self.assertTrue(self.observer.added_recieved)

        # Reset flags:
        self.observer.removed_recieved = False
        self.observer.added_recieved = False

        # Setting a parent when there is one set; both signals should be fired:
        parent = object()
        self.childmodel.parent = parent

        self.assertTrue(self.observer.removed_recieved)
        self.assertTrue(self.observer.added_recieved)

        # Reset flags:
        self.observer.removed_recieved = False
        self.observer.added_recieved = False

        # Setting an identical parent does not fire anything:
        self.childmodel.parent = parent

        self.assertFalse(self.observer.removed_recieved)
        self.assertFalse(self.observer.added_recieved)

        # Reset flags:
        self.observer.removed_recieved = False
        self.observer.added_recieved = False

        # Setting the parent to None when it was set; should only fire the removed signal:
        self.childmodel.parent = None

        self.assertTrue(self.observer.removed_recieved)
        self.assertFalse(self.observer.added_recieved)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename # @UnresolvedImport

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from pyxrd.generic.views import BaseView

class EditInSituBehaviourView(BaseView):
    builder = resource_filename(__name__, "glade/edit_insitu_behaviour.glade")
    top = "edit_insitu_behaviour"

    widget_format = "behaviour_%s"
    parameter_table = "edit_insitu_behaviour"

    def __init__(self, meta, **kwargs):
        assert (meta is not None), "EditInSituBehaviourView needs a model's Meta class!"
        BaseView.__init__(self, **kwargs)

        self.props = [ prop for prop in meta.all_properties if getattr(prop, "visible", False) ]

        def create_label(prop):
            new_lbl = Gtk.Label(label=prop.text)
            new_lbl.set_use_markup(True)
            return new_lbl

        def create_input(prop):
            new_inp = self.add_widget(prop)
            new_inp.set_name(self.widget_format % prop.label)
            return new_inp

        self.create_input_table(
            self[self.parameter_table], self.props, 
            num_columns=1, widget_callbacks = [
                create_label,
                create_input
            ]
        )

    pass #end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from mvc import Controller

from pyxrd.generic.controllers import BaseController

from pyxrd.phases.models.CSDS import CSDS_distribution_types

class EditCSDSTypeController(BaseController):
    """ 
        Controller for the selection of the type of CSDS Model
    """
    auto_adapt = False

    distributions_controller = None

    def reset_type_store(self):
        if self.view is not None:
            combo = self.view["cmb_type"]
            store = Gtk.ListStore(str, object) # @UndefinedVariable

            for cls in CSDS_distribution_types:
                store.append([cls.Meta.description, cls])
            combo.set_model(store)

            for row in store:
                if type(self.model.CSDS_distribution) == store.get_value(row.iter, 1):
                    combo.set_active_iter(row.iter)
                    break
            return store

    def register_view(self, view):
        self.view = view
        combo = self.view["cmb_type"]
        combo.connect('changed', self.on_changed)
        cell = Gtk.CellRendererText() # @UndefinedVariable
        combo.pack_start(cell, True)
        combo.add_attribute(cell, 'markup', 0)
        self.reset_type_store()
        self.reset_distributions_controller()

    @BaseController.model.setter
    def _set_model(self, model):
        super(EditCSDSTypeController, self)._set_model(model)
        self.reset_distributions_controller()


    def reset_distributions_controller(self):
        if self.view is not None:
            if self.distributions_controller is None:
                self.distributions_controller = EditCSDSDistributionController(
                    model=self.model.CSDS_distribution,
                    view=self.view,
                    parent=self)
            else:
                self.distributions_controller.model = self.model.CSDS_distribution

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_changed(self, combo, user_data=None):
        itr = combo.get_active_iter()
        if itr is not None:
            cls = combo.get_model().get_value(itr, 1)
            if not type(self.model.CSDS_distribution) == cls:
                new_csds_model = cls(parent=self.model)
                self.model.CSDS_distribution = new_csds_model
                self.distributions_controller.model = new_csds_model

    pass # end of class

class EditCSDSDistributionController(BaseController):
    """ 
        Controller for the CSDS Models 
        Handles the creation of widgets based on the property descriptor settings
    """

    auto_adapt = False

    def reset_view(self):
        if self.view is not None:
            self.view.reset_params()
            for prop in self.model.Meta.all_properties:
                if getattr(prop, "refinable", False):
                    self.view.add_param_widget(
                        self.view.widget_format % prop.label, prop.label,
                        prop.minimum, prop.maximum
                    )
            self.view.update_figure(self.model.distrib[0])
            self.register_adapters()
            self.adapt()

    def register_view(self, view):
        if self.model is not None:
            self.reset_view()

    @BaseController.model.setter
    def model(self, model):
        super(EditCSDSDistributionController, self)._set_model(model)
        self.reset_view()

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @Controller.observe("updated", signal=True)
    def notif_updated(self, model, prop_name, info):
        if self.model.distrib is not None and not self.model.phase.project.before_needs_update_lock:
            try: self.view.update_figure(self.model.distrib[0])
            except BaseException as error:
                logger.exception("Caught unhandled exception: %s" % error)

    pass # end of class



#!/usr/bin/python
# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os
from pyxrd.data import settings

from pyxrd.project.models import Project
from pyxrd.phases.models import Component, Phase

def generate_expandables(
            filename_format, phase_name, maxR,
            phase_kwargs_AD, phase_kwargs_EG, phase_kwargs_350,
            code_AD, code_EG, code_350,
            comp_kwargs_AD, comp_kwargs_EG, comp_kwargs_350):
        """
            Generates a list of phase descriptions for a combination of an
            AD, EG and 350° Ca-saturated phase linked together
        """
        return [
            ('%s' + (filename_format % R), [
                    (dict(R=R, name=phase_name + (' R%d Ca-AD' % R), **phase_kwargs_AD), code_AD, comp_kwargs_AD),
                    (dict(R=R, name=phase_name + (' R%d Ca-EG' % R), based_on=phase_name + (' R%d Ca-AD' % R), **phase_kwargs_EG), code_EG, comp_kwargs_EG),
                    (dict(R=R, name=phase_name + (' R%d Ca-350' % R), based_on=phase_name + (' R%d Ca-AD' % R), **phase_kwargs_350), code_350, comp_kwargs_350)
            ]) for R in range(maxR)
        ]

def run(args=None, ui_callback=None):

    """ 
    How this script works:
    
     - 'code_length' is the length of the aliases keys (see below)
     - 'aliases' is a dictionary contain 4-character long keys describing a
        specific layer-type (or with other words: a Component object)
        E.g. dS2w stands for Di-octahedral Smectite with 2 layers of water.
        The values are file path formats, in which a single '%s' string placeholder
        will be filled with the absolute path to the default components folder.  
     - 'default_phases' is an initially empty list that will be filled with two-
        tuples. The first element in this tuple is the filename of the generated
        phases, the second element is describing what this phase contains. This
        second element is again a tuple, containing three parts:
            - A dictionary of key-word arguments passed on to the Phase
              constructor. If a 'based_on' keyword is defined, an attempt is
              made to translate it to an earlier generated phase. This way, it
              is possible to pass the name of an earlier generated phase, and
              the script will pass in the actual Phase object instead.
            - A component code (string) built by the keys of the 'aliases'
              dictionary. This string's length should be a multiple of 'code_length'.
              There is no separator, rather, the 'code_length' is used to split the
              code into its parts.   
            - Component keyword arguments dictionaries: this is a dictionary in
              which the keys match with the components code parts. The values are
              property-value dictionaries used to set Component properties after
              importing them. Similarly to the Phases' 'based_on' keyword, the
              value for the 'linked_with' key is translated to the actual
              Component named as such.   
     
    ### Setup:
    """
    code_length = 4
    aliases = {
        'C   ': '%sChlorite.cmp',
        'K   ': '%sKaolinite.cmp',
        'I   ': '%sIllite.cmp',
        'Se  ': '%sSerpentine.cmp',
        'T   ': '%sTalc.cmp',
        'Ma  ': '%sMargarite.cmp',
        'Pa  ': '%sParagonite.cmp',
        'L   ': '%sLeucophyllite.cmp',
        'dS2w': '%sDi-Smectite/Di-Smectite - Ca 2WAT.cmp',
        'dS1w': '%sDi-Smectite/Di-Smectite - Ca 1WAT.cmp',
        'dS0w': '%sDi-Smectite/Di-Smectite - Ca Dehydr.cmp',
        'dS2g': '%sDi-Smectite/Di-Smectite - Ca 2GLY.cmp',
        'dS1g': '%sDi-Smectite/Di-Smectite - Ca 1GLY.cmp',
        'dSht': '%sDi-Smectite/Di-Smectite - Ca Heated.cmp',
        'tS2w': '%sTri-Smectite/Tri-Smectite - Ca 2WAT.cmp',
        'tS1w': '%sTri-Smectite/Tri-Smectite - Ca 1WAT.cmp',
        'tS0w': '%sTri-Smectite/Tri-Smectite - Ca Dehydr.cmp',
        'tS2g': '%sTri-Smectite/Tri-Smectite - Ca 2GLY.cmp',
        'tS1g': '%sTri-Smectite/Tri-Smectite - Ca 1GLY.cmp',
        'tSht': '%sTri-Smectite/Tri-Smectite - Ca Heated.cmp',
        'dV2w': '%sDi-Vermiculite/Di-Vermiculite - Ca 2WAT.cmp',
        'dV1w': '%sDi-Vermiculite/Di-Vermiculite - Ca 1WAT.cmp',
        'dV0w': '%sDi-Vermiculite/Di-Vermiculite - Ca Dehydr.cmp',
        'dV2g': '%sDi-Vermiculite/Di-Vermiculite - Ca 2GLY.cmp',
        'dV1g': '%sDi-Vermiculite/Di-Vermiculite - Ca 1GLY.cmp',
        'dVht': '%sDi-Vermiculite/Di-Vermiculite - Ca Heated.cmp',
    }
    default_phases = []

    """ 
    ### Commonly used inherit flag dicts:
    """
    inherit_S = dict(
        inherit_ucp_a=True,
        inherit_ucp_b=True,
        inherit_delta_c=True,
        inherit_layer_atoms=True,
    )

    inherit_all = dict(
        inherit_d001=True,
        inherit_default_c=True,
        inherit_interlayer_atoms=True,
        inherit_atom_relations=True,
        **inherit_S
    )

    inherit_phase = dict(
        inherit_display_color=True,
        inherit_sigma_star=True,
        inherit_CSDS_distribution=True,
        inherit_probabilities=True
    )

    """ 
    ### Single-layer phases:
    """
    default_phases += [
        ('%sKaolinite.phs', [(dict(R=0, name='Kaolinite'), 'K   ', {}), ]),
        ('%sIllite.phs', [(dict(R=0, name='Illite'), 'I   ', {})]),
        ('%sSerpentine.phs', [(dict(R=0, name='Serpentine'), 'Se  ', {})]),
        ('%sTalc.phs', [(dict(R=0, name='Talc'), 'T   ', {})]),
        ('%sChlorite.phs', [(dict(R=0, name='Chlorite'), 'C   ', {})]),
        ('%sMargarite.phs', [(dict(R=0, name='Margarite'), 'Ma  ', {})]),
        ('%sLeucophyllite.phs', [(dict(R=0, name='Leucophyllite'), 'L   ', {})]),
        ('%sParagonite.phs', [(dict(R=0, name='Paragonite'), 'Pa  ', {})]),
    ]

    """      
    ### Dioctahedral smectites:
    """
    S_code_AD = 'dS2w'
    S_code_EG = 'dS2g'
    S_code_350 = 'dSht'
    S_inh_comp_args = {
        'dS2g': dict(linked_with='dS2w', **inherit_S),
        'dSht': dict(linked_with='dS2w', **inherit_S),
    }

    SS_code_AD = S_code_AD + 'dS1w'
    SS_code_EG = S_code_EG + 'dS1g'
    SS_code_350 = S_code_350 + 'dS1g'
    SS_inh_comp_args = dict(S_inh_comp_args)
    SS_inh_comp_args.update({
        'dS1g': dict(linked_with='dS1w', **inherit_S),
    })


    SSS_code_AD = SS_code_AD + 'dS0w'
    SSS_code_EG = SS_code_EG + 'dS0w'
    SSS_code_350 = SS_code_350 + 'dS0w'
    SSS_inh_comp_args = dict(SS_inh_comp_args)
    SSS_inh_comp_args.update({
        'dS0w': dict(linked_with='dS0w', **inherit_S),
    })

    default_phases += [
        ('%sSmectites/Di-Smectite Ca.phs', [
                (dict(R=0, name='S R0 Ca-AD'), S_code_AD, {}),
                (dict(R=0, name='S R0 Ca-EG', based_on='S R0 Ca-AD', **inherit_phase), S_code_EG, S_inh_comp_args),
                (dict(R=0, name='S R0 Ca-350', based_on='S R0 Ca-AD', **inherit_phase), S_code_350, S_inh_comp_args)
        ]),
    ]
    default_phases += generate_expandables(
        'Smectites/SS/Di-SS R%d Ca.phs', 'SS', 4,
        {}, inherit_phase, inherit_phase,
        SS_code_AD, SS_code_EG, SS_code_350,
        {}, SS_inh_comp_args, SS_inh_comp_args,
    )
    default_phases += generate_expandables(
        'Smectites/SSS/Di-SSS R%d Ca.phs', 'SSS', 3,
        {}, inherit_phase, inherit_phase,
        SSS_code_AD, SSS_code_EG, SSS_code_350,
        {}, SSS_inh_comp_args, SSS_inh_comp_args,
    )

    """      
    ### Trioctahedral smectites:
    """
    tS_code_AD = 'tS2w'
    tS_code_EG = 'tS2g'
    tS_code_350 = 'tSht'
    tS_inh_comp_args = {
        'tS2g': dict(linked_with='tS2w', **inherit_S),
        'tSht': dict(linked_with='tS2w', **inherit_S),
    }

    tSS_code_AD = tS_code_AD + 'tS1w'
    tSS_code_EG = tS_code_EG + 'tS1g'
    tSS_code_350 = tS_code_350 + 'tS1g'
    tSS_inh_comp_args = dict(S_inh_comp_args)
    tSS_inh_comp_args.update({
        'tS1g': dict(linked_with='tS1w', **inherit_S),
    })

    tSSS_code_AD = tSS_code_AD + 'tS0w'
    tSSS_code_EG = tSS_code_EG + 'tS0w'
    tSSS_code_350 = tSS_code_350 + 'tS0w'
    tSSS_inh_comp_args = dict(SS_inh_comp_args)
    tSSS_inh_comp_args.update({
        'tS0w': dict(linked_with='tS0w', **inherit_S),
    })

    default_phases += [
        ('%sSmectites/Tri-Smectite Ca.phs', [
                (dict(R=0, name='S R0 Ca-AD'), tS_code_AD, {}),
                (dict(R=0, name='S R0 Ca-EG', based_on='S R0 Ca-AD', **inherit_phase), tS_code_EG, tS_inh_comp_args),
                (dict(R=0, name='S R0 Ca-350', based_on='S R0 Ca-AD', **inherit_phase), tS_code_350, tS_inh_comp_args)
        ]),
    ]
    default_phases += generate_expandables(
        'Smectites/SS/Tri-SS R%d Ca.phs', 'SS', 4,
        {}, inherit_phase, inherit_phase,
        tSS_code_AD, tSS_code_EG, tSS_code_350,
        {}, tSS_inh_comp_args, tSS_inh_comp_args,
    )
    default_phases += generate_expandables(
        'Smectites/SSS/Tri-SSS R%d Ca.phs', 'SSS', 3,
        {}, inherit_phase, inherit_phase,
        tSSS_code_AD, tSSS_code_EG, tSSS_code_350,
        {}, tSSS_inh_comp_args, tSSS_inh_comp_args,
    )

    """      
    ### Dioctahedral vermiculites:
    """
    V_code_AD = 'dV2w'
    V_code_EG = 'dV2g'
    V_code_350 = 'dVht'
    V_inh_comp_args = {
        'dV2g': dict(linked_with='dV2w', **inherit_S),
        'dVht': dict(linked_with='dV2w', **inherit_S),
    }

    VV_code_AD = V_code_AD + 'dV1w'
    VV_code_EG = V_code_EG + 'dV1g'
    VV_code_350 = V_code_350 + 'dV1g'
    VV_inh_comp_args = dict(V_inh_comp_args)
    VV_inh_comp_args.update({
        'dV1g': dict(linked_with='dV1w', **inherit_S),
    })


    VVV_code_AD = VV_code_AD + 'dV0w'
    VVV_code_EG = VV_code_EG + 'dV0w'
    VVV_code_350 = VV_code_350 + 'dV0w'
    VVV_inh_comp_args = dict(VV_inh_comp_args)
    VVV_inh_comp_args.update({
        'dV0w': dict(linked_with='dV0w', **inherit_S),
    })

    default_phases += [
        ('%sVermiculites/Di-Vermiculite Ca.phs', [
                (dict(R=0, name='V R0 Ca-AD'), V_code_AD, {}),
                (dict(R=0, name='V R0 Ca-EG', based_on='V R0 Ca-AD', **inherit_phase), V_code_EG, V_inh_comp_args),
                (dict(R=0, name='V R0 Ca-350', based_on='V R0 Ca-AD', **inherit_phase), V_code_350, V_inh_comp_args)
        ]),
    ]
    default_phases += generate_expandables(
        'Vermiculites/VV/Di-VV R%d Ca.phs', 'VV', 4,
        {}, inherit_phase, inherit_phase,
        VV_code_AD, VV_code_EG, VV_code_350,
        {}, VV_inh_comp_args, VV_inh_comp_args,
    )
    default_phases += generate_expandables(
        'Vermiculites/VVV/Di-VVV R%d Ca.phs', 'VVV', 3,
        {}, inherit_phase, inherit_phase,
        VVV_code_AD, VVV_code_EG, VVV_code_350,
        {}, VVV_inh_comp_args, VVV_inh_comp_args,
    )

    """      
    ### Kaolinite - Smectites:
    """
    K_code = 'K   '
    K_inh_comp_args = {
        'K   ': dict(linked_with='K   ', **inherit_all),
    }

    KS_code_AD = K_code + S_code_AD
    KS_code_EG = K_code + S_code_EG
    KS_code_350 = K_code + S_code_350
    KS_inh_comp_args = dict(S_inh_comp_args)
    KS_inh_comp_args.update(K_inh_comp_args)

    KSS_code_AD = K_code + SS_code_AD
    KSS_code_EG = K_code + SS_code_EG
    KSS_code_350 = K_code + SS_code_350
    KSS_inh_comp_args = dict(SS_inh_comp_args)
    KSS_inh_comp_args.update(K_inh_comp_args)

    KSSS_code_AD = K_code + SSS_code_AD
    KSSS_code_EG = K_code + SSS_code_EG
    KSSS_code_350 = K_code + SSS_code_350
    KSSS_inh_comp_args = dict(SSS_inh_comp_args)
    KSSS_inh_comp_args.update(K_inh_comp_args)

    default_phases += generate_expandables(
        'Kaolinite-Smectites/KS/KS R%d Ca.phs', 'KS', 4,
        {}, inherit_phase, inherit_phase,
        KS_code_AD, KS_code_EG, KS_code_350,
        {}, KS_inh_comp_args, KS_inh_comp_args,
    )
    default_phases += generate_expandables(
        'Kaolinite-Smectites/KSS/KSS R%d Ca.phs', 'KSS', 3,
        {}, inherit_phase, inherit_phase,
        KSS_code_AD, KSS_code_EG, KSS_code_350,
        {}, KSS_inh_comp_args, KSS_inh_comp_args,
    )
    default_phases += generate_expandables(
        'Kaolinite-Smectites/KSSS/KSSS R%d Ca.phs', 'KSSS', 2,
        {}, inherit_phase, inherit_phase,
        KSSS_code_AD, KSSS_code_EG, KSSS_code_350,
        {}, KSSS_inh_comp_args, KSSS_inh_comp_args,
    )

    """       
    ### Illite - Smectites:
    """
    I_code = 'I   '
    I_inh_comp_args = {
        'I   ': dict(linked_with='I   ', **inherit_all),
    }

    IS_code_AD = I_code + S_code_AD
    IS_code_EG = I_code + S_code_EG
    IS_code_350 = I_code + S_code_350
    IS_inh_comp_args = dict(S_inh_comp_args)
    IS_inh_comp_args.update(I_inh_comp_args)

    ISS_code_AD = I_code + SS_code_AD
    ISS_code_EG = I_code + SS_code_EG
    ISS_code_350 = I_code + SS_code_350
    ISS_inh_comp_args = dict(SS_inh_comp_args)
    ISS_inh_comp_args.update(I_inh_comp_args)

    ISSS_code_AD = I_code + SSS_code_AD
    ISSS_code_EG = I_code + SSS_code_EG
    ISSS_code_350 = I_code + SSS_code_350
    ISSS_inh_comp_args = dict(SSS_inh_comp_args)
    ISSS_inh_comp_args.update(I_inh_comp_args)

    default_phases += generate_expandables(
        'Illite-Smectites/IS/IS R%d Ca.phs', 'IS', 4,
        {}, inherit_phase, inherit_phase,
        IS_code_AD, IS_code_EG, IS_code_350,
        {}, IS_inh_comp_args, IS_inh_comp_args,
    )
    default_phases += generate_expandables(
        'Illite-Smectites/ISS/ISS R%d Ca.phs', 'ISS', 3,
        {}, inherit_phase, inherit_phase,
        ISS_code_AD, ISS_code_EG, ISS_code_350,
        {}, ISS_inh_comp_args, ISS_inh_comp_args,
    )
    default_phases += generate_expandables(
        'Illite-Smectites/ISSS/ISSS R%d Ca.phs', 'ISSS', 2,
        {}, inherit_phase, inherit_phase,
        ISSS_code_AD, ISSS_code_EG, ISSS_code_350,
        {}, ISSS_inh_comp_args, ISSS_inh_comp_args,
    )

    """        
    ### Chlorite - Smectites:
    """
    C_code = 'C   '
    C_inh_comp_args = {
        'C   ': dict(linked_with='C   ', **inherit_all),
    }

    CS_code_AD = C_code + tS_code_AD
    CS_code_EG = C_code + tS_code_EG
    CS_code_350 = C_code + tS_code_350
    CS_inh_comp_args = dict(tS_inh_comp_args)
    CS_inh_comp_args.update(C_inh_comp_args)

    CSS_code_AD = C_code + tSS_code_AD
    CSS_code_EG = C_code + tSS_code_EG
    CSS_code_350 = C_code + tSS_code_350
    CSS_inh_comp_args = dict(tSS_inh_comp_args)
    CSS_inh_comp_args.update(C_inh_comp_args)

    CSSS_code_AD = C_code + tSSS_code_AD
    CSSS_code_EG = C_code + tSSS_code_EG
    CSSS_code_350 = C_code + tSSS_code_350
    CSSS_inh_comp_args = dict(tSSS_inh_comp_args)
    CSSS_inh_comp_args.update(C_inh_comp_args)

    default_phases += generate_expandables(
        'Chlorite-Smectites/CS/CS R%d Ca.phs', 'CS', 4,
        {}, inherit_phase, inherit_phase,
        CS_code_AD, CS_code_EG, CS_code_350,
        {}, CS_inh_comp_args, CS_inh_comp_args,
    )
    default_phases += generate_expandables(
        'Chlorite-Smectites/CSS/CSS R%d Ca.phs', 'CSS', 3,
        {}, inherit_phase, inherit_phase,
        CSS_code_AD, CSS_code_EG, CSS_code_350,
        {}, CSS_inh_comp_args, CSS_inh_comp_args,
    )
    default_phases += generate_expandables(
        'Chlorite-Smectites/CSSS/CSSS R%d Ca.phs', 'CSSS', 2,
        {}, inherit_phase, inherit_phase,
        CSSS_code_AD, CSSS_code_EG, CSSS_code_350,
        {}, CSSS_inh_comp_args, CSSS_inh_comp_args,
    )

    """      
    ### Talc - Smectites:
    """
    T_code = 'T   '
    T_inh_comp_args = {
        'T   ': dict(linked_with='T   ', **inherit_all),
    }

    TS_code_AD = T_code + S_code_AD
    TS_code_EG = T_code + S_code_EG
    TS_code_350 = T_code + S_code_350
    TS_inh_comp_args = dict(S_inh_comp_args)
    TS_inh_comp_args.update(T_inh_comp_args)

    TSS_code_AD = T_code + SS_code_AD
    TSS_code_EG = T_code + SS_code_EG
    TSS_code_350 = T_code + SS_code_350
    TSS_inh_comp_args = dict(SS_inh_comp_args)
    TSS_inh_comp_args.update(T_inh_comp_args)

    TSSS_code_AD = T_code + SSS_code_AD
    TSSS_code_EG = T_code + SSS_code_EG
    TSSS_code_350 = T_code + SSS_code_350
    TSSS_inh_comp_args = dict(SSS_inh_comp_args)
    TSSS_inh_comp_args.update(T_inh_comp_args)

    default_phases += generate_expandables(
        'Talc-Smectites/TS/TS R%d Ca.phs', 'TS', 4,
        {}, inherit_phase, inherit_phase,
        TS_code_AD, TS_code_EG, TS_code_350,
        {}, TS_inh_comp_args, TS_inh_comp_args,
    )
    default_phases += generate_expandables(
        'Talc-Smectites/TSS/TSS R%d Ca.phs', 'TSS', 3,
        {}, inherit_phase, inherit_phase,
        TSS_code_AD, TSS_code_EG, TSS_code_350,
        {}, TSS_inh_comp_args, TSS_inh_comp_args,
    )
    default_phases += generate_expandables(
        'Talc-Smectites/TSSS/TSSS R%d Ca.phs', 'TSSS', 2,
        {}, inherit_phase, inherit_phase,
        TSSS_code_AD, TSSS_code_EG, TSSS_code_350,
        {}, TSSS_inh_comp_args, TSSS_inh_comp_args,
    )

    """        
    ### Illite - Chlorite - Smectites:
    """
    IC_code = I_code + C_code
    IC_inh_comp_args = dict(I_inh_comp_args)
    IC_inh_comp_args.update(C_inh_comp_args)

    ICS_code_AD = IC_code + S_code_AD
    ICS_code_EG = IC_code + S_code_EG
    ICS_inh_comp_args = dict(S_inh_comp_args)
    ICS_inh_comp_args.update(IC_inh_comp_args)

    ICSS_code_AD = IC_code + SS_code_AD
    ICSS_code_EG = IC_code + SS_code_EG
    ICSS_inh_comp_args = dict(SS_inh_comp_args)
    ICSS_inh_comp_args.update(IC_inh_comp_args)

    ICSSS_code_AD = IC_code + SSS_code_AD
    ICSSS_code_EG = IC_code + SSS_code_EG
    ICSSS_inh_comp_args = dict(SSS_inh_comp_args)
    ICSSS_inh_comp_args.update(IC_inh_comp_args)

    default_phases += [
        ('%sIllite-Chlorite-Smectites/ICS/ICS R0 Ca.phs', [
                (dict(R=0, name='ICS R0 Ca-AD'), ICS_code_AD, {}),
                (dict(R=0, name='ICS R0 Ca-EG', based_on='ICS R0 Ca-AD', **inherit_phase), ICS_code_EG, ICS_inh_comp_args)
        ]),
        ('%sIllite-Chlorite-Smectites/ICS/ICS R1 Ca.phs', [
                (dict(R=1, name='ICS R1 Ca-AD'), ICS_code_AD, {}),
                (dict(R=1, name='ICS R1 Ca-EG', based_on='ICS R1 Ca-AD', **inherit_phase), ICS_code_EG, ICS_inh_comp_args)
        ]),
        ('%sIllite-Chlorite-Smectites/ICS/ICS R2 Ca.phs', [
                (dict(R=2, name='ICS R2 Ca-AD'), ICS_code_AD, {}),
                (dict(R=2, name='ICS R2 Ca-EG', based_on='ICS R2 Ca-AD', **inherit_phase), ICS_code_EG, ICS_inh_comp_args)
        ]),

        ('%sIllite-Chlorite-Smectites/ICSS/ICSS R0 Ca.phs', [
                (dict(R=0, name='ICSS R0 Ca-AD'), ICSS_code_AD, {}),
                (dict(R=0, name='ICSS R0 Ca-EG', based_on='ICSS R0 Ca-AD', **inherit_phase), ICSS_code_EG, ICSS_inh_comp_args)
        ]),
        ('%sIllite-Chlorite-Smectites/ICSS/ICSS R1 Ca.phs', [
                (dict(R=1, name='ICSS R1 Ca-AD'), ICSS_code_AD, {}),
                (dict(R=1, name='ICSS R1 Ca-EG', based_on='ICSS R1 Ca-AD', **inherit_phase), ICSS_code_EG, ICSS_inh_comp_args)
        ]),

        ('%sIllite-Chlorite-Smectites/ICSSS/ICSSS R0 Ca.phs', [
                (dict(R=0, name='ICSSS R0 Ca-AD'), ICSSS_code_AD, {}),
                (dict(R=0, name='ICSSS R0 Ca-EG', based_on='ICSSS R0 Ca-AD', **inherit_phase), ICSSS_code_EG, ICSSS_inh_comp_args)
        ]),
    ]

    """        
    ### Kaolinite - Chlorite - Smectites:
    """
    KC_code = K_code + C_code
    KC_inh_comp_args = dict(K_inh_comp_args)
    KC_inh_comp_args.update(C_inh_comp_args)

    KCS_code_AD = KC_code + S_code_AD
    KCS_code_EG = KC_code + S_code_EG
    KCS_inh_comp_args = dict(S_inh_comp_args)
    KCS_inh_comp_args.update(KC_inh_comp_args)

    KCSS_code_AD = KC_code + SS_code_AD
    KCSS_code_EG = KC_code + SS_code_EG
    KCSS_inh_comp_args = dict(SS_inh_comp_args)
    KCSS_inh_comp_args.update(KC_inh_comp_args)

    KCSSS_code_AD = KC_code + SSS_code_AD
    KCSSS_code_EG = KC_code + SSS_code_EG
    KCSSS_inh_comp_args = dict(SSS_inh_comp_args)
    KCSSS_inh_comp_args.update(KC_inh_comp_args)

    default_phases += [
        ('%sKaolinite-Chlorite-Smectites/KCS/KCS R0 Ca.phs', [
                (dict(R=0, name='KCS R0 Ca-AD'), KCS_code_AD, {}),
                (dict(R=0, name='KCS R0 Ca-EG', based_on='KCS R0 Ca-AD', **inherit_phase), KCS_code_EG, KCS_inh_comp_args)
        ]),
        ('%sKaolinite-Chlorite-Smectites/KCS/KCS R1 Ca.phs', [
                (dict(R=1, name='KCS R1 Ca-AD'), KCS_code_AD, {}),
                (dict(R=1, name='KCS R1 Ca-EG', based_on='KCS R1 Ca-AD', **inherit_phase), KCS_code_EG, KCS_inh_comp_args)
        ]),
        ('%sKaolinite-Chlorite-Smectites/KCS/KCS R2 Ca.phs', [
                (dict(R=2, name='KCS R2 Ca-AD'), KCS_code_AD, {}),
                (dict(R=2, name='KCS R2 Ca-EG', based_on='KCS R2 Ca-AD', **inherit_phase), KCS_code_EG, KCS_inh_comp_args)
        ]),

        ('%sKaolinite-Chlorite-Smectites/KCSS/KCSS R0 Ca.phs', [
                (dict(R=0, name='KCSS R0 Ca-AD'), KCSS_code_AD, {}),
                (dict(R=0, name='KCSS R0 Ca-EG', based_on='KCSS R0 Ca-AD', **inherit_phase), KCSS_code_EG, KCSS_inh_comp_args)
        ]),
        ('%sKaolinite-Chlorite-Smectites/KCSS/KCSS R1 Ca.phs', [
                (dict(R=1, name='KCSS R1 Ca-AD'), KCSS_code_AD, {}),
                (dict(R=1, name='KCSS R1 Ca-EG', based_on='KCSS R1 Ca-AD', **inherit_phase), KCSS_code_EG, KCSS_inh_comp_args)
        ]),

        ('%sKaolinite-Chlorite-Smectites/KCSSS/KCSSS R0 Ca.phs', [
                (dict(R=0, name='KCSSS R0 Ca-AD'), KCSSS_code_AD, {}),
                (dict(R=0, name='KCSSS R0 Ca-EG', based_on='KCSSS R0 Ca-AD', **inherit_phase), KCSSS_code_EG, KCSSS_inh_comp_args)
        ]),
    ]

    """
    ### Actual object generation routine:
    """
    import queue
    import threading

    def ioworker(in_queue, stop):
        """
            Saves Phase objects from the in_queue.
            If the Queue is empty this function will only stop
            if the 'stop' event is set.
        """
        while True:
            try:
                phases_path, phases = in_queue.get(timeout=0.5)
                create_dir_recursive(phases_path)
                Phase.save_phases(phases, phases_path)
                in_queue.task_done()
            except queue.Empty:
                if not stop.is_set():
                    continue
                else:
                    return

    save_queue = queue.Queue()
    io_stop = threading.Event()
    iothread = threading.Thread(target=ioworker, args=(save_queue, io_stop))
    iothread.start()

    def phaseworker(in_queue, save_queue, stop):
        """
            Parses Phase descriptions into actual objects and passes them
            to the save_queue.
            'stop' should be a threading.Event() that should be toggled
            once all elements have been Queued.
            This way, the worker will only stop once the Queue is really empty,
            and not when it's processing faster than the Queue can be filled. 
        """
        while True:
            try:
                phases_path, phase_descr = in_queue.get(timeout=0.5)
                project = Project()
                phase_lookup = {}
                component_lookup = {}

                for phase_kwargs, code, comp_props in phase_descr:

                    # create phase:
                    G = len(code) / code_length
                    based_on = None
                    if "based_on" in phase_kwargs:
                        based_on = phase_lookup.get(phase_kwargs.pop("based_on"), None)
                    phase = Phase(G=G, parent=project, **phase_kwargs)
                    phase.based_on = based_on
                    phase_lookup[phase.name] = phase

                    # derive upper and lower limits for the codes using code lengths:
                    limits = list(zip(
                        list(range(0, len(code), code_length)),
                        list(range(code_length, len(code) + 1, code_length))
                    ))

                    # create components:
                    phase.components[:] = []
                    for ll, ul in limits:
                        part = code[ll: ul]
                        for component in Component.load_components(aliases[part] % (settings.DATA_REG.get_directory_path("DEFAULT_COMPONENTS") + "/"), parent=phase):
                            component.resolve_json_references()
                            phase.components.append(component)
                            props = comp_props.get(part, {})
                            for prop, value in props.items():
                                if prop == 'linked_with':
                                    value = component_lookup[value]
                                setattr(component, prop, value)

                            component_lookup[part] = component

                # put phases on the save queue:
                phases_path = phases_path % (settings.DATA_REG.get_directory_path("DEFAULT_PHASES") + "/")
                save_queue.put((phases_path, list(phase_lookup.values())))
                # Flag this as finished
                in_queue.task_done()
            except queue.Empty:
                if not stop.is_set():
                    continue
                else:
                    return

    phase_queue = queue.Queue()
    phase_stop = threading.Event()
    phasethread = threading.Thread(target=phaseworker, args=(phase_queue, save_queue, phase_stop))
    phasethread.start()

    # Queue phases:
    for phases_path, phase_descr in default_phases:
        phase_queue.put((phases_path, phase_descr))

    # Signal phaseworker it can stop if the phase_queue is emptied:
    phase_stop.set()
    while phasethread.is_alive():
        # Try to join the thread, but don't block, inform the UI
        # of our progress if a callback is provided:
        phasethread.join(timeout=0.1)
        if callable(ui_callback):
            progress = float(len(default_phases) - phase_queue.qsize()) / float(len(default_phases))
            ui_callback(progress)

    if callable(ui_callback):
        ui_callback(1.0)

    # Signal the IO worker the phaseworker has stopped, so it can stop
    # if the save_queue is empty
    io_stop.set()
    while iothread.is_alive():
        # Try to join the thread, but don't block
        iothread.join(timeout=0.1)

    pass # end of run

def create_dir_recursive(path):
    """
        Creates the path 'path' recursively.
    """
    to_create = []
    while not os.path.exists(path):
        to_create.insert(0, path)
        path = os.path.dirname(path)
    for path in to_create[:-1]:
        os.mkdir(path)



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .objectliststore_controllers import (
    TreeViewMixin,
    TreeModelMixin,
    TreeControllerMixin,
    ObjectListStoreController,
    ChildObjectListStoreController,
    InlineObjectListStoreController,
)
from .base_controller import BaseController
from .dialog_controller import DialogController

__all__ = [
    "TreeViewMixin",
    "TreeModelMixin",
    "TreeControllerMixin",
    "ObjectListStoreController",
    "ChildObjectListStoreController",
    "InlineObjectListStoreController",
    "BaseController",
    "DialogController"
]



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .data_mixin import DataMixin
from .refinable_mixin import RefinableMixin

__ALL__ = [
    "DataMixin",
    "RefinableMixin"
]


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pyxrd.data import settings

from pyxrd.calculations.CSDS import calculate_distribution
from pyxrd.calculations.data_objects import CSDSData
from pyxrd.generic.models import DataModel
from pyxrd.generic.io import storables, Storable

from pyxrd.refinement.refinables.mixins import RefinementGroup, RefinementValue
from pyxrd.refinement.refinables.metaclasses import PyXRDRefinableMeta
from pyxrd.refinement.refinables.properties import DataMixin, RefinableMixin

from mvc.models.properties import (
    GetActionMixin, LabeledProperty, BoolProperty, FloatProperty,
    ReadOnlyMixin, SetActionMixin
)
from mvc.models.properties.signal_mixin import SignalMixin

class _AbstractCSDSDistribution(DataModel, Storable, metaclass=PyXRDRefinableMeta):

    # MODEL INTEL:
    class Meta(DataModel.Meta):
        description = "Abstract CSDS distr."
        explanation = ""

    phase = property(DataModel.parent.fget, DataModel.parent.fset)

    # PROPERTIES:
    _data_object = None
    @property
    def data_object(self):
        return self._data_object

    inherited = BoolProperty(
        default=False, text="Inherited",
        visible=False, persistent=False,
        signal_name="data_changed", mix_with=(SignalMixin,)
    )

    distrib = LabeledProperty(
        default=None, text="CSDS Distribution",
        tabular=True, visible=False, persistent=False,
        get_action_name="_update_distribution",
        signal_name="data_changed", 
        mix_with=(SignalMixin, GetActionMixin,)
    )

    # PROPERTIES:
    #: The maximum value of this distribution
    maximum = FloatProperty(
        default=0.0, text="Maximum CSDS",
        minimum=1, maximum=1000,
        tabular=True, persistent=False, visible=False,
        mix_with=(ReadOnlyMixin, DataMixin)
    )

    #: The minimum value of this distribution
    minimum = FloatProperty(
        default=0.0, text="Maximum CSDS",
        minimum=1, maximum=1000,
        tabular=True, persistent=False, visible=False,
        mix_with=(ReadOnlyMixin, DataMixin)
    )

    average = FloatProperty(
        default=0.0, text="Average CSDS",
        minimum=1, maximum=200,
        tabular=True, persistent=True, visible=True, refinable=True,
        signal_name="data_changed", 
        set_action_name="_update_distribution",
        mix_with=(SignalMixin, DataMixin, RefinableMixin, SetActionMixin)
    )

    alpha_scale = FloatProperty(
        default=0.0, text="α scale factor",
        minimum=0.0, maximum=10.0,
        tabular=True, persistent=True, visible=True, refinable=True,
        signal_name="data_changed", 
        set_action_name="_update_distribution",
        mix_with=(SignalMixin, DataMixin, RefinableMixin, SetActionMixin)
    )

    alpha_offset = FloatProperty(
        default=0.0, text="α offset factor",
        minimum=-5, maximum=5,
        tabular=True, persistent=True, visible=True, refinable=True,
        signal_name="data_changed", 
        set_action_name="_update_distribution",
        mix_with=(SignalMixin, DataMixin, RefinableMixin, SetActionMixin)
    )

    beta_scale = FloatProperty(
        default=0.0, text="β² scale factor",
        minimum=0.0, maximum=10.0,
        tabular=True, persistent=True, visible=True, refinable=True,
        signal_name="data_changed", 
        set_action_name="_update_distribution",
        mix_with=(SignalMixin, DataMixin, RefinableMixin, SetActionMixin)
    )

    beta_offset = FloatProperty(
        default=0.0, text="β² offset factor",
        minimum=-5, maximum=5,
        tabular=True, persistent=True, visible=True, refinable=True,
        signal_name="data_changed", 
        set_action_name="_update_distribution",
        mix_with=(SignalMixin, DataMixin, RefinableMixin, SetActionMixin)
    )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, average=10, alpha_scale=0.9485, alpha_offset=-0.0017,
            beta_scale=0.1032, beta_offset=0.0034, *args, **kwargs):

        super(_AbstractCSDSDistribution, self).__init__(*args, **kwargs)

        self._data_object = CSDSData()

        type(self).average._set(self, average)
        type(self).maximum._set(self, int(settings.LOG_NORMAL_MAX_CSDS_FACTOR * average))
        type(self).minimum._set(self, 1)
        type(self).alpha_scale._set(self, alpha_scale)
        type(self).alpha_offset._set(self, alpha_offset)
        type(self).beta_scale._set(self, beta_scale)
        type(self).beta_offset._set(self, beta_offset)

        self._update_distribution()

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def _update_distribution(self):
        type(self).maximum._set(self, int(settings.LOG_NORMAL_MAX_CSDS_FACTOR * self.average))
        self._distrib = calculate_distribution(self.data_object)

    pass # end of class

@storables.register()
class LogNormalCSDSDistribution(_AbstractCSDSDistribution, RefinementGroup):

    # MODEL INTEL:
    class Meta(_AbstractCSDSDistribution.Meta):
        description = "Generic log-normal CSDS distr. (Eberl et al. 1990)"
        store_id = "LogNormalCSDSDistribution"

    # REFINEMENT GROUP IMPLEMENTATION:
    @property
    def refine_title(self):
        return "CSDS Distribution"

    @property
    def refine_descriptor_data(self):
        return dict(
            phase_name=self.phase.name,
            component_name="*"
        )

    pass # end of class

@storables.register()
class DritsCSDSDistribution(_AbstractCSDSDistribution, RefinementValue):

    # MODEL INTEL:
    class Meta(_AbstractCSDSDistribution.Meta):
        description = "Log-normal CSDS distr. (Drits et. al, 1997)"
        store_id = "DritsCSDSDistribution"

    # PROPERTIES:
    alpha_scale = FloatProperty(
        default=0.9485, text="α scale factor",
        minimum=0.0, maximum=10.0,
        tabular=True, persistent=False, visible=False, refinable=False,
        mix_with=(ReadOnlyMixin, DataMixin, RefinableMixin)
    )

    alpha_offset = FloatProperty(
        default=0.017, text="α offset factor",
        minimum=-5, maximum=5,
        tabular=True, persistent=False, visible=False, refinable=False,
        mix_with=(ReadOnlyMixin, DataMixin, RefinableMixin)
    )

    beta_scale = FloatProperty(
        default=0.1032, text="β² scale factor",
        minimum=0.0, maximum=10.0,
        tabular=True, persistent=False, visible=False, refinable=False,
        mix_with=(ReadOnlyMixin, DataMixin, RefinableMixin)
    )

    beta_offset = FloatProperty(
        default=0.0034, text="β² offset factor",
        minimum=-5, maximum=5,
        tabular=True, persistent=False, visible=False, refinable=False,
        mix_with=(ReadOnlyMixin, DataMixin, RefinableMixin)
    )

    # REFINEMENT VALUE IMPLEMENTATION:
    @property
    def refine_title(self):
        return "Average CSDS"

    @property
    def refine_descriptor_data(self):
        return dict(
            phase_name=self.phase.name,
            component_name="*",
            property_name=self.refine_title
        )

    @property
    def refine_value(self):
        return self.average
    @refine_value.setter
    def refine_value(self, value):
        self.average = value

    @property
    def refine_info(self):
        return self.average_ref_info

    @property
    def is_refinable(self):
        return not self.inherited

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        for key in ["alpha_scale", "alpha_offset", "beta_scale", "beta_offset"]:
            kwargs.pop(key, None)
        super(DritsCSDSDistribution, self).__init__(*args, **kwargs)

    pass # end of class

CSDS_distribution_types = [
    LogNormalCSDSDistribution,
    DritsCSDSDistribution
]



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import zipfile

from pyxrd.generic.io import get_case_insensitive_glob, COMPRESSION

from pyxrd.file_parsers.json_parser import JSONParser

from .namespace import atom_type_parsers

@atom_type_parsers.register_parser()
class JSONAtomTypeParser(JSONParser):
    """
        Atomic scattering factors JSON file parser
    """

    description = "Atom types JSON file *.ZTL"
    extensions = get_case_insensitive_glob("*.ZTL")
    mimetypes = ["application/octet-stream", "application/zip"]

    __file_mode__ = "rb"

    @classmethod
    def _parse_header(cls, filename, fp, data_objects=None, close=False):
        return data_objects # just pass it on, nothing to do

    @classmethod
    def write(cls, filename, items, props):
        """
            Saves multiple AtomTypes to a single (JSON) file.
        """
        with zipfile.ZipFile(filename, 'w', compression=COMPRESSION) as zfile:
            for i, atom_type in enumerate(items):
                zfile.writestr("%d###%s" % (i, atom_type.uuid), atom_type.dump_object())

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os, struct
from io import SEEK_SET #, SEEK_CUR, SEEK_END

import numpy as np

from pyxrd.generic.io.utils import get_case_insensitive_glob
from pyxrd.generic.utils import u

from ..base_parser import BaseParser
from .namespace import xrd_parsers
from .xrd_parser_mixin import XRDParserMixin


def cap(lower, value, upper, out=None):
    if value < lower or value > upper:
        return out if out is not None else min(max(value, lower), upper)
    else:
        return value

@xrd_parsers.register_parser()
class RDParser(XRDParserMixin, BaseParser):
    """
        Philips Binary V3 & V5 *.RD format parser
    """

    description = "Phillips Binary V3/V5 *.RD"
    extensions = get_case_insensitive_glob("*.RD")
    mimetypes = ["application/octet-stream", ]

    __file_mode__ = "rb"

    @classmethod
    def _parse_header(cls, filename, fp, data_objects=None, close=False):
        f = fp

        try:
            basename = u(os.path.basename(filename))
        except:
            basename = None

        # Adapt XRDFile list
        data_objects = cls._adapt_data_object_list(data_objects, num_samples=1)

        # Go to the start of the file
        f.seek(0, SEEK_SET)

        # Read file format version:
        version = f.read(2).decode()

        if version in ("V3", "V5"):

            # Read diffractometer, target and focus type:
            f.seek(84, SEEK_SET)
            diffractomer_type, target_type, focus_type = struct.unpack("bbb", f.read(3))
            diffractomer_type = {
                0: b"PW1800",
                1: b"PW1710 based system",
                2: b"PW1840",
                3: b"PW3710 based system",
                4: b"Undefined",
                5: b"X'Pert MPD"
            }[cap(0, diffractomer_type, 5, 4)]
            target_type = {
                0: b"Cu",
                1: b"Mo",
                2: b"Fe",
                3: b"Cr",
                4: b"Other"
            }[cap(0, target_type, 3, 4)]
            focus_type = {
                0: b"BF",
                1: b"NF",
                2: b"FF",
                3: b"LFF",
                4: b"Unkown",
            }[cap(0, focus_type, 3, 4)]

            # Read wavelength information:
            f.seek(94, SEEK_SET)
            alpha1, alpha2, alpha_factor = struct.unpack("ddd", f.read(24))
            # Read sample name:
            f.seek(146, SEEK_SET)
            sample_name = u(f.read(16).replace(b"\0", b""))

            # Read data limits:
            f.seek(214)
            twotheta_step, twotheta_min, twotheta_max = struct.unpack("ddd", f.read(24))
            twotheta_count = int((twotheta_max - twotheta_min) / twotheta_step)

            # Set data start:
            data_start = {
                "V3": 250,
                "V5": 810
            }[version]

            data_objects[0].update(
                filename=basename,
                name=sample_name,
                twotheta_min=twotheta_min,
                twotheta_max=twotheta_max,
                twotheta_step=twotheta_step,
                twotheta_count=twotheta_count,
                target_type=target_type,
                alpha1=alpha1,
                alpha2=alpha2,
                alpha_factor=alpha_factor,
                data_start=data_start,
                version=version
            )

        else:
            raise IOError("Only V3 and V5 *.RD files are supported!")

        if close: f.close()
        return data_objects

    @classmethod
    def _parse_data(cls, filename, fp, data_objects=None, close=False):
        f = fp

        # RD files are singletons, so no need to iterate over the list,
        # there is only one XRDFile instance:
        if data_objects[0].data == None:
            data_objects[0].data = []

        # Parse data:
        if f is not None:
            if data_objects[0].version in ("V3", "V5"):
                # Move to start of data:
                f.seek(data_objects[0].data_start)
                n = 0
                while n < data_objects[0].twotheta_count:
                    y, = struct.unpack("H", f.read(2))
                    data_objects[0].data.append([
                        data_objects[0].twotheta_min + data_objects[0].twotheta_step * float(n + 0.5),
                        float(y)
                    ])
                    n += 1
            else:
                raise IOError("Only V3 and V5 *.RD files are supported!")

        data_objects[0].data = np.array(data_objects[0].data)

        if close: f.close()
        return data_objects

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from mvc.controller import Controller

from .status_bar_mixin import StatusBarMixin

class BaseController(StatusBarMixin, Controller):
    file_filters = ("All Files", "*.*")
    widget_handlers = {} # handlers can be string representations of a class method

    pass # end of class
